/*
 * Copyright (c) 2001-2003 Swedish Institute of Computer Science.
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 *
 * This file is part of the lwIP TCP/IP stack.
 * 
 * Author: Adam Dunkels <adam@sics.se>
 *
 */

#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <time.h>

/* ============================================== */
#include <net/ethernet.h>
#include <linux/if_packet.h>
#include <sys/socket.h>
/* ============================================== */

#include "lwip/opt.h"

#include "lwip/init.h"

#include "lwip/mem.h"
#include "lwip/memp.h"
#include "lwip/sys.h"
#include "lwip/timers.h"

#include "lwip/ip_addr.h"

#include "lwip/dns.h"
#include "lwip/dhcp.h"

#include "lwip/stats.h"

#include "lwip/tcp.h"
#include "lwip/inet_chksum.h"

#include "netif/etharp.h"

#include "lwip/tcpip.h"
/* #include "lwip/sockets.h" */

#include "netif/tapif.h"
#include "netif/tunif.h"

#include "netif/unixif.h"
#include "netif/dropif.h"
#include "netif/pcapif.h"

#include "netif/tcpdump.h"

#if LWIP_HAVE_SLIPIF
#include "netif/slipif.h"
#define SLIP_PTY_TEST 1
#endif

#if PPP_SUPPORT
#include "netif/ppp/pppos.h"
#define PPP_PTY_TEST 1
#include <termios.h>
#endif

#include "lwip/ip_addr.h"
#include "arch/perf.h"

#include "lwip/apps/httpd.h"
#include "apps/udpecho/udpecho.h"
#include "apps/tcpecho/tcpecho.h"
#include "apps/shell/shell.h"
#include "apps/chargen/chargen.h"
#include "apps/netio/netio.h"
#include "lwip/apps/netbiosns.h"
#include "lwip/apps/sntp.h"
#include "lwip/apps/snmp.h"
#include "lwip/apps/snmp_mib2.h"
#include "apps/snmp_private_mib/private_mib.h"

#if LWIP_RAW
#include "lwip/icmp.h"
#include "lwip/raw.h"
#include "ports/unix/proj/lib/lwipopts.h"
#endif


/* =========================================================== */
int socket_raw;
unsigned char mac[6] = {0x08, 0x00, 0x27, 0x1c, 0xdb, 0x5c}; /* TODO */
int interface_index = 2;  /* TODO */
/* =========================================================== */

/* Forward declarations. */
err_t netif_tst_setup(struct netif *netif);



/* (manual) host IP configuration */
static ip_addr_t ipaddr, netmask, gw;

struct netif netif_tst;

/* nonstatic debug cmd option, exported in lwipopts.h */
unsigned char debug_flags;

/** @todo add options for selecting netif, starting DHCP client etc */
static struct option longopts[] = {
  /* turn on debugging output (if build with LWIP_DEBUG) */
  {"debug", no_argument,        NULL, 'd'},
  /* help */
  {"help", no_argument, NULL, 'h'},
#if LWIP_IPV4
  /* gateway address */
  {"gateway", required_argument, NULL, 'g'},
  /* ip address */
  {"ipaddr", required_argument, NULL, 'i'},
  /* netmask */
  {"netmask", required_argument, NULL, 'm'},
  /* ping destination */
  {"ping",   required_argument, NULL, 'p'},
#endif /* LWIP_IPV4 */
  /* new command line options go here! */
  {NULL,   0,                 NULL,  0}
};
#define NUM_OPTS ((sizeof(longopts) / sizeof(struct option)) - 1)

static void init_netifs(void);

static void usage(void)
{
  unsigned char i;

  printf("options:\n");
  for (i = 0; i < NUM_OPTS; i++) {
    printf("-%c --%s\n",longopts[i].val, longopts[i].name);
  }
}

static err_t low_level_output(struct netif *netif, struct pbuf *p)
{
  char buf[1514];
  /* ssize_t written; */

  LWIP_ASSERT("chek", netif == &netif_tst);
  /* initiate transfer(); */
  pbuf_copy_partial(p, buf, p->tot_len, 0);

  if (send(socket_raw, buf, p->tot_len, 0) > 0 ) {
    return ERR_OK;
  } else {
    printf("send - %s. Socket=%d\n", strerror(errno), socket_raw);
    return ERR_IF;
  }
}

static struct pbuf *low_level_input(struct netif *netif)
{
  struct pbuf *p;
  u16_t len;
  char buf[4*1514];
  struct eth_hdr *ethhdr;


  /* Obtain the size of the packet and put it into the "len"
     variable. */
  len = recv(socket_raw, buf, sizeof(buf), 0);
  if (len == (u16_t)-1) {
    perror("read returned -1");
    exit(1);
  }

  /* We allocate a pbuf chain of pbufs from the pool. */
  p = pbuf_alloc(PBUF_RAW, len, PBUF_POOL);
  if (p != NULL) {
    pbuf_take(p, buf, len);
    /* acknowledge that packet has been read(); */
    
    ethhdr = (struct eth_hdr *)p->payload;
    switch(lwip_htons(ethhdr->type)) {
    case ETHTYPE_IP:
    case ETHTYPE_ARP:
        if (netif->input(p, netif) != ERR_OK) {
            pbuf_free(p);
            printf("Pacote descartado\n");
        }
        break;
    default:
        pbuf_free(p);
        printf("Tipo invalido de eth-header: %d. Pacote descartado", lwip_htons(ethhdr->type));
        break;
    }    
  } else {
    /* drop packet(); */
    printf("input: could not allocate pbuf\n");
  }

  return p;
}

err_t netif_tst_setup (struct netif *netif) {

  struct sockaddr_ll linkLevelInfo;

  socket_raw = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
  if (socket_raw < 0) {
      printf("Open socket error=%s\n\n", strerror(errno));
      exit(-1);
  }

  /* bind to eth:    */
  memset(&linkLevelInfo, 0, sizeof(linkLevelInfo));
  linkLevelInfo.sll_family = AF_PACKET;
  linkLevelInfo.sll_ifindex =  interface_index;
  linkLevelInfo.sll_protocol = htons(ETH_P_ALL);

  linkLevelInfo.sll_halen = ETH_ALEN;
  memcpy (linkLevelInfo.sll_addr, mac, ETH_ALEN);

  if (bind(socket_raw, (struct sockaddr*)(&linkLevelInfo), sizeof(linkLevelInfo)) != 0){
    printf("Bind socket error\n\n");
    exit(-1);
  }

  netif->name[0] = 'e';
  netif->name[1] = '0';
  netif->output = etharp_output;
  netif->linkoutput = low_level_output;
  netif->mtu = 1500;
  netif->hwaddr_len = 6;
  netif->flags = NETIF_FLAG_BROADCAST | NETIF_FLAG_ETHARP | NETIF_FLAG_IGMP | NETIF_FLAG_LINK_UP;
  memcpy (netif->hwaddr, mac, sizeof(mac));

  return ERR_OK;
}

/*-----------------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------------*/

static void
init_netifs(void)
{
  netif_add(&netif_tst, ip_2_ip4(&ipaddr), ip_2_ip4(&netmask), ip_2_ip4(&gw), NULL, 
            netif_tst_setup, tcpip_input);
  netif_set_default(&netif_tst);
  netif_set_up(&netif_tst);

  tcpecho_init();
  /*  sys_timeout(5000, tcp_debug_timeout, NULL);*/
}

static void
tcpip_init_done(void *arg)
{
  sys_sem_t *sem;
  sem = (sys_sem_t *)arg;

  init_netifs();
 
  sys_sem_signal(sem);
}

/*-----------------------------------------------------------------------------------*/
static void
main_thread(void *arg)
{
  sys_sem_t sem;
  LWIP_UNUSED_ARG(arg);

  if(sys_sem_new(&sem, 0) != ERR_OK) {
    LWIP_ASSERT("Failed to create semaphore", 0);
  }
  tcpip_init(tcpip_init_done, &sem);
  sys_sem_wait(&sem);
  printf("TCP/IP initialized.\n\n\n");


  while (1) {
    low_level_input(&netif_tst);
  }
}
/*-----------------------------------------------------------------------------------*/
int
main(int argc, char **argv)
{
  int ch;

  /* startup defaults (may be overridden by one or more opts) */
  IP_ADDR4(&gw,      10, 0,  2,  1);   /* TODO */
  IP_ADDR4(&netmask, 255,255,255,0);   /* TODO */
  IP_ADDR4(&ipaddr,  10, 0,  2,  121); /* TODO */
  
  /* use debug flags defined by debug.h */
  debug_flags = LWIP_DBG_OFF;
  
  while ((ch = getopt_long(argc, argv, "dhg:i:m:p:", longopts, NULL)) != -1) {
    switch (ch) {
      case 'd':
        debug_flags |= (LWIP_DBG_ON|LWIP_DBG_TRACE|LWIP_DBG_STATE|LWIP_DBG_FRESH|LWIP_DBG_HALT);
        break;
      case 'h':
        usage();
        exit(0);
        break;
#if LWIP_IPV4
      case 'g':
        ipaddr_aton(optarg, &gw);
        break;
      case 'i':
        ipaddr_aton(optarg, &ipaddr);
        break;
      case 'm':
        ipaddr_aton(optarg, &netmask);
        break;
#endif /* LWIP_IPV4 */
      default:
        usage();
        break;
    }
  }
  argc -= optind;
  argv += optind;
  
   
  sys_thread_new("main_thread", main_thread, NULL, DEFAULT_THREAD_STACKSIZE, DEFAULT_THREAD_PRIO);
  pause();
  return 0;
}
/*-----------------------------------------------------------------------------------*/
